#include "memory_incremental.h"
#include "memory_preallocated.h"
#include "page_queue_partitioned.h"
#include "pipe_paging.h"



PipePaging::PipePaging(Memory *memory, offset_t size)
: _memory(NULL), _size(size)
{
    /* Reserve space for two pipe regions. */

    _memory = new MemoryPreallocated(memory, size * 2);
    _queue = new PageQueuePartitioned();
    _pages = new Pages(_memory, _queue);

    for (unsigned int i = 0; i < 2; i++)
        _regions[i] = NULL;
}

PipePaging::PipePaging(offset_t size)
: _size(size)
{
    /* Reserve space for two pipe regions. */

    _memory_base = new MemoryIncremental();
    _memory = new MemoryPreallocated(_memory_base, size * 2);
    _queue = new PageQueuePartitioned();
    _pages = new Pages(_memory, _queue);

    for (unsigned int i = 0; i < 2; i++)
        _regions[i] = NULL;
}

/* Detach one endpoint. */

void PipePaging::detach()
{
    if (!_endpoints)
        return;
    else
        _endpoints--;

    /* Return if the other endpoint is attached. */

    if (_endpoints)
        return;

    /* Discard all regions from the pipe. */

    for (unsigned int i = 0; i < 2; i++)
    {
        PageMapper *mapper = _regions[i];

        if (mapper != NULL)
        {
            mapper->detach();
            _regions[i] = NULL;
            delete mapper;
        }
    }

    /* Delete the page collection and related objects. */

    delete _pages;
    delete _queue;
    delete _memory;

    /* Delete any underlying memory object. */

    if (_memory_base != NULL)
        delete _memory_base;
}

/* Add a region to the sequence. */

PageMapper *PipePaging::add_region()
{
    /* If the writer already accesses a different region to the reader, no new
       region is added. */

    if (_writing != _reading)
        return NULL;

    /* Select the other region of the pair being maintained. */

    _writing = 1 - _writing;

    /* Make a new mapper for the region. */

    PageMapper *mapper = new PageMapper(&_accessors[_writing], _pages);

    /* Initialise and record the mapper. */

    mapper->attach();
    mapper->set_data_size(0);

    _regions[_writing] = mapper;
    return mapper;
}

/* Return the current region for reading. */

PageMapper *PipePaging::current_region()
{
    return _regions[_reading];
}

/* Return the next region for the reader if the writer is using a different one.
   Otherwise, return NULL. */

PageMapper *PipePaging::next_region()
{
    /* If the reader already accesses the same region to the writer, no next
       region can be obtained. */

    if (_reading == _writing)
        return NULL;

    /* Detach and discard the current page mapper. */

    PageMapper *mapper = _regions[_reading];

    if (mapper != NULL)
    {
        mapper->detach();
        _regions[_reading] = NULL;
        delete mapper;
    }

    /* Return the next region. */

    _reading = 1 - _reading;
    return _regions[_reading];
}

// vim: tabstop=4 expandtab shiftwidth=4
